import eve2cml.main


def test_centered_line():
    # Test when data is None
    full = eve2cml.main.centered_line_with_stars()
    assert len(full) == 80

    partial = eve2cml.main.centered_line_with_stars("this")
    assert len(full) == 80
    assert " this " in partial



import argparse
import io
import logging
import sys
import xml.etree.ElementTree as ET
import zipfile
from pathlib import Path
from typing import List

import yaml

from ._version import __version__
from .eve import Lab, Network, Node, Objects, Topology
from .log import initialize_logging
from .mapper import Eve2CMLmapper

_LOGGER = logging.getLogger(__name__)


def parse_xml(xml_content: str, filename: str, mapper: Eve2CMLmapper):
    lab = ET.fromstring(xml_content)
    lab_name = lab.attrib.get("name", "")
    lab_version = lab.attrib.get("version", "")
    script_timeout = int(lab.attrib.get("scripttimeout", 0))
    countdown = int(lab.attrib.get("countdown", 0))
    lock = int(lab.attrib.get("lock", 0))
    sat = int(lab.attrib.get("sat", 0))

    desc = lab.find(".//description")
    task = lab.find(".//body")

    description = ""
    if desc is not None and desc.text:
        description += f"## Description: \n\n{desc.text}\n\n"
    if task is not None and task.text:
        description += f"## Task: \n\n{task.text}\n\n"
    description += f"Imported from {filename} via `eve2cml` converter"

    parsed_lab = Lab(
        name=lab_name,
        version=lab_version,
        scripttimeout=script_timeout,
        countdown=countdown,
        lock=bool(lock),
        sat=sat,
        description=description,
        topology=Topology(nodes=Node.parse(lab), networks=Network.parse(lab)),
        objects=Objects.parse(lab, ".//objects", filename),
        filename=filename,
        mapper=mapper,
    )

    return parsed_lab


def convert_file(content: str, filename: str, mapper: Eve2CMLmapper) -> Lab:
    _LOGGER.info("Parse XML file %s", filename)
    lab = parse_xml(content, filename, mapper)
    _LOGGER.info("Done with file %s", filename)
    return lab


def convert_files(file_or_zip: str, mapper: Eve2CMLmapper) -> List[Lab]:
    lab: List[Lab] = []
    if zipfile.is_zipfile(file_or_zip):
        with zipfile.ZipFile(file_or_zip, "r") as zip_file:
            for file_info in zip_file.infolist():
                dirname = str(Path(file_info.filename).parent)
                if dirname.startswith("__MACOSX"):
                    continue
                filename = Path(file_info.filename).name
                if filename.endswith(".unl"):
                    try:
                        content = zip_file.read(file_info.filename)
                        dir = f"{dirname}--{filename}" if dirname != "." else filename
                        lab.append(convert_file(content.decode("utf-8"), dir, mapper))
                    except KeyError:
                        print(f"File {filename} not found in the ZIP archive.")
    else:
        try:
            with open(file_or_zip, encoding="utf-8") as xml_file:
                txt_content = xml_file.read()
                lab.append(convert_file(txt_content, file_or_zip, mapper))
        except FileNotFoundError as exc:
            _LOGGER.critical("%s", exc)
            sys.exit(1)
    return lab


def dump_as_text(out: io.TextIOWrapper, lab: Lab, dump_all: bool):
    out.write(">>> Nodes <<<\n")
    for node in lab.topology.nodes:
        out.write(f"{node}\n")
        num_ifaces = len(node.interfaces) - 1
        for idx, interface in enumerate(node.interfaces):
            bullet = "|--" if idx < num_ifaces else "\\__"
            out.write(f"  {bullet} {interface}\n")
        out.write("\n")

    out.write(">>> Networks <<<\n")
    for network in lab.topology.networks:
        out.write(f"{network}")
    out.write("\n")

    out.write(">>> Text objects <<<\n")
    for text_object in lab.objects.textobjects:
        out.write(f"{text_object}")
    out.write("\n")

    if not dump_all:
        return

    out.write(">>> Tasks <<<\n")
    for task in lab.objects.tasks:
        out.write(f"{task}")
    out.write("\n")

    out.write(">>> Configs <<<\n")
    for config in lab.objects.configs:
        out.write(f"{config}")
    out.write("\n")

    out.write(">>> Config sets <<<\n")
    for configset in lab.objects.configsets:
        out.write(f"Config Set ID: {configset.id}\n")
        out.write(f"Config Set Name: {configset.name}\n")
        out.write("Contained Configs:\n")
        for config in configset.configs:
            out.write(f"{config}\n")
        out.write("\n")


def centered_line_with_stars(name="", cols=80) -> str:
    if len(name) == 0:
        return "*" * cols
    num_asterisks = cols - len(name) + 2
    asterisks_each_side = num_asterisks // 2
    asterisks_left = "*" * asterisks_each_side
    asterisks_right = "*" * (
        asterisks_each_side if num_asterisks % 2 == 0 else asterisks_each_side + 1
    )
    return f"{asterisks_left} {name} {asterisks_right}"


def main():
    parser = argparse.ArgumentParser(
        description="Convert UNL/XML topologies to CML2 topologies"
    )
    parser.epilog = f"Example: {parser.prog} exportedlabs.zip"
    parser.add_argument(
        "-V", "--version", action="version", version=f"%(prog)s {__version__}"
    )

    parser.add_argument(
        "--level",
        default="warning",
        choices=["debug", "info", "warning", "error", "critical"],
        help="specify the log level, default is warning",
    )
    parser.add_argument(
        "--stdout", action="store_true", help="do not store in files, print to stdout"
    )
    parser.add_argument("--nocolor", action="store_true", help="no color log output")
    parser.add_argument("--dump", action="store_true", help="Dump the mapper as YAML")
    parser.add_argument("--mapper", help="custom mapper YAML file")
    parser.add_argument("-t", "--text", action="store_true", help="text output")
    parser.add_argument(
        "--all", action="store_true", help="print all objects in text mode"
    )
    parser.add_argument(
        "file_or_zip", nargs="+", help="Path to either a UNL or  ZIP with UNL file"
    )
    args = parser.parse_args()

    initialize_logging(args.level, args.nocolor)

    if args.dump:
        _LOGGER.warning("dumping the mapper into %s", args.file_or_zip)
        filename = sys.stdout.fileno() if args.stdout else args.file_or_zip[0]
        with open(filename, "w") as fh:
            Eve2CMLmapper().load().dump(fh)
        return

    if args.all and not args.text:
        _LOGGER.warn("--all is only relevant with text output, ignoring")

    mapper = Eve2CMLmapper().load(args.mapper)
    labs: List[Lab] = []
    for arg in args.file_or_zip:
        labs.extend(convert_files(arg, mapper))

    # YAML is the default
    if not args.text:

        def yaml_multiline_string_pipe(dumper, data):
            text_list = [line.rstrip() for line in data.splitlines()]
            fixed_data = "\n".join(text_list)
            if len(text_list) > 1:
                return dumper.represent_scalar(
                    "tag:yaml.org,2002:str", fixed_data, style="|"
                )
            return dumper.represent_scalar("tag:yaml.org,2002:str", fixed_data)

        yaml.add_representer(str, yaml_multiline_string_pipe)

        if args.stdout:
            for lab in labs:
                print(
                    centered_line_with_stars(
                        str(Path(lab.filename).with_suffix(".yaml"))
                    )
                )
                sys.stdout.write(yaml.dump(lab.as_cml_dict(), sort_keys=False))
                print(centered_line_with_stars())
            return

        for lab in labs:
            cml_filename = Path(lab.filename).with_suffix(".yaml")
            with open(cml_filename, "w", encoding="utf-8") as cml_file:
                cml_file.write(yaml.dump(lab.as_cml_dict(), sort_keys=False))
        return

    # this is simply text output
    for lab in labs:
        txt_filename = (
            sys.stdout.fileno()
            if args.stdout
            else str(Path(lab.filename).with_suffix(".txt"))
        )
        with open(txt_filename, "w", encoding="utf-8") as out:
            dump_as_text(out, lab, args.all)
    return



from typing import List

from .network import Network
from .node import Node


class Topology:
    def __init__(self, nodes: List[Node], networks: List[Network]):
        self.nodes = nodes or []
        self.networks = networks or []

    def __str__(self):
        return f"Nodes: {self.nodes}, Networks: {self.networks}"

    def next_node_id(self) -> int:
        return max([node.id for node in self.nodes]) + 1



from typing import List
from xml.etree.ElementTree import Element


class Network:
    def __init__(self, id: int, obj_type: str, name: str, top: int, left: int):
        self.id = id
        self.obj_type = obj_type
        self.name = name
        self.left = left
        self.top = top
        # ignored for the moment
        self.style = "Solid"
        self.linkstyle = "Straight"
        self.color = ""
        self.label = ""
        self.visibility = "0"
        self.icon = "lan.png"

    def __str__(self):
        return f"ID: {self.id}, Name: {self.name}, Type: {self.obj_type}"

    @classmethod
    def parse(cls, lab: Element) -> List["Network"]:
        networks: List[Network] = []
        for network_elem in lab.findall(".//networks/network"):
            network = Network(
                id=int(network_elem.attrib.get("id", 0)),
                obj_type=network_elem.attrib.get("type", "unknown"),
                name=network_elem.attrib.get("name", ""),
                top=int(network_elem.attrib.get("top", 0)),
                left=int(network_elem.attrib.get("left", 0)),
            )
            networks.append(network)
        return networks



from typing import List
from xml.etree.ElementTree import Element

from .config import Config


class ConfigSet:
    def __init__(self, id: int, name: str, configs: List[Config]):
        self.id = id
        self.name = name
        self.configs = configs

    def __str__(self):
        return f"Config Set ID: {self.id}, Name: {self.name}, Contained Configs: {self.configs}"

    @classmethod
    def parse(cls, lab: Element, path) -> List["ConfigSet"]:
        configsets: List[ConfigSet] = []
        for configset_elem in lab.findall(path):
            configset = ConfigSet(
                id=int(configset_elem.attrib.get("id", 0)),
                name=configset_elem.attrib.get("name", "unknown"),
                configs=Config.parse(configset_elem, "config"),
            )
            configsets.append(configset)
        return configsets



from .config import Config
from .configset import ConfigSet
from .interface import Interface
from .lab import Lab
from .network import Network
from .node import Node
from .objects import Objects
from .task import Task
from .textobject import TextObject
from .topology import Topology

__all__ = [
    "Config",
    "ConfigSet",
    "Interface",
    "Lab",
    "Network",
    "Node",
    "Objects",
    "TextObject",
    "Task",
    "Topology",
]



import logging
from typing import Any, Dict, List, Tuple

from ..mapper import Eve2CMLmapper
from .interface import Interface
from .network import Network
from .node import Node
from .objects import Objects
from .topology import Topology

_LOGGER = logging.getLogger(__name__)


class CMLlink:
    def __init__(
        self,
        from_id: int,
        from_slot: int,
        to_id: int,
        to_slot: int,
        label: str,
    ):
        self.from_id = from_id
        self.from_slot = from_slot
        self.to_id = to_id
        self.to_slot = to_slot
        self.label = label

    def as_cml_dict(self, idx: int):
        return {
            "id": f"l{idx}",
            "n1": f"n{self.from_id}",
            "i1": f"i{self.from_slot}",
            "n2": f"n{self.to_id}",
            "i2": f"i{self.to_slot}",
            "label": f"{self.label}-{idx}",
            "conditioning": {},
        }


class Lab:
    def __init__(
        self,
        name: str,
        version: str,
        scripttimeout: int,
        countdown: int,
        lock: bool,
        sat: int,
        description: str,
        topology: Topology,
        objects: Objects,
        filename: str,
        mapper: Eve2CMLmapper,
    ):
        self.name = name
        self.version = version
        self.scripttimeout = scripttimeout
        self.countdown = countdown
        self.lock = lock
        self.sat = sat
        self.description = description
        self.topology = topology
        self.objects = objects

        # conversion data
        self.mapper = mapper
        self.filename = filename

        self._links: List[Dict[str, Any]] = []
        self._current_link_id = 0

    def as_cml_dict(self):
        result: Dict[str, Any] = {}
        result["lab"] = {
            "notes": self.description,
            "description": f"Imported from {self.filename} via eve2cml converter",
            "title": self.name,
            "version": "0.1.0",
        }
        result["annotations"] = self.objects.cml_annotations()
        result["links"] = self.cml_links()
        result["nodes"] = [
            node.as_cml_dict(node.id, self) for node in self.topology.nodes
        ]

        labels = {node["label"] for node in result["nodes"]}
        if len(labels) != len(result["nodes"]):
            _LOGGER.warning(
                "node labels are not unique, this can not be imported into CML!"
            )

        return result

    def __str__(self):
        return f"Lab: {self.name}, Version: {self.version}, Script Timeout: {self.scripttimeout}, Countdown: {self.countdown}, Lock: {self.lock}, SAT: {self.sat}"

    def _network_ifaces(self, network_id: int) -> List[Interface]:
        interfaces: List[Interface] = []
        for node in self.topology.nodes:
            for iface in node.interfaces:
                if iface.network_id == network_id:
                    interfaces.append(iface)
        return interfaces

    def _connect_internal_network(
        self, ums_node_id: int, network_id: int, name: str
    ) -> List[CMLlink]:
        found_ids: List[Tuple[int, int]] = []
        for node in self.topology.nodes:
            for iface in node.interfaces:
                if iface.network_id == network_id:
                    found_ids.append((int(node.id), iface.slot))
        return [
            CMLlink(ums_node_id, idx, *found, name)
            for idx, found in enumerate(found_ids)
        ]

    def _insert_ext_conn(self, network: Network, config: str, offset=0) -> Node:
        next_node_id = self.topology.next_node_id()
        obj_type = "cml_ext_conn"
        ext_conn = Node(
            id=next_node_id,
            name=f"ext-{network.obj_type}-{network.name}",
            interfaces=[
                Interface(
                    id=0,
                    obj_type=obj_type,
                    name="port",
                    network_id=network.id,
                    slot=0,
                )
            ],
            obj_type=obj_type,
            template=obj_type,
            left=network.left,
            top=network.top - offset,
            ethernet=1,
        )
        ext_conn.cml_config = config
        self.topology.nodes.append(ext_conn)
        return ext_conn

    def _insert_ums(self, network: Network, num_ifaces: int):
        _LOGGER.info("ums")
        next_node_id = self.topology.next_node_id()
        ums_links = self._connect_internal_network(
            next_node_id, network.id, network.name
        )

        obj_type = "cml_ums"
        ums = Node(
            id=next_node_id,
            name=f"ums-{network.obj_type}-{network.name}",
            interfaces=[
                Interface(
                    id=idx,
                    name=f"port{idx}",
                    obj_type="ethernet",
                    slot=idx,
                    network_id=network.id,
                )
                for idx in range(num_ifaces)
            ],
            obj_type=obj_type,
            template=obj_type,
            left=network.left,
            top=network.top,
        )
        ums.ethernet = 8 if num_ifaces < 8 else num_ifaces
        self.topology.nodes.append(ums)
        for link in ums_links:
            self._links.append(link.as_cml_dict(self._current_link_id))
            self._current_link_id += 1

    def cml_links(self) -> List[Dict[str, Any]]:
        for network in self.topology.networks:
            _LOGGER.info("Processing network %d, %s", network.id, network.name)
            ifcelist = self._network_ifaces(network.id)
            num_ifaces = len(ifcelist)

            if network.obj_type == "bridge":
                if num_ifaces == 2:
                    _LOGGER.info("p2p")
                    from_iface = ifcelist[0]
                    to_iface = ifcelist[1]
                    self._links.append(
                        CMLlink(
                            from_iface.node_id,
                            from_iface.slot,
                            to_iface.node_id,
                            to_iface.slot,
                            network.name,
                        ).as_cml_dict(self._current_link_id)
                    )
                    self._current_link_id += 1
                elif num_ifaces > 2:
                    self._insert_ums(network, num_ifaces)
                else:
                    _LOGGER.error("Can't deal with bridge with %d ifaces", num_ifaces)

            elif network.obj_type.startswith("nat"):
                _LOGGER.info("nat")
                if num_ifaces != 1:
                    _LOGGER.error("NAT interface has %d ifaces", num_ifaces)
                    continue
                ext_conn = self._insert_ext_conn(network, config="nat")
                self._links.append(
                    CMLlink(
                        ifcelist[0].node_id,
                        ifcelist[0].slot,
                        ext_conn.id,
                        0,
                        network.name,
                    ).as_cml_dict(self._current_link_id)
                )
                self._current_link_id += 1

            elif network.obj_type.startswith("pnet"):
                _LOGGER.info("pnet")
                bridge_number = int(network.obj_type.lstrip("pnet"))
                ext_conn = self._insert_ext_conn(
                    network, config=f"bridge{bridge_number}", offset=64
                )
                self._insert_ums(network, num_ifaces + 1)

            elif network.obj_type == "internal":
                _LOGGER.warning("Ignoring internal network %s", network.name)

            else:
                _LOGGER.error(
                    "Unhandled network type %s (%d port(s)) in %s",
                    network.obj_type,
                    num_ifaces,
                    self.filename,
                )

        return self._links



import logging
from pathlib import Path

import pytest
from eve2cml import mapper


def test_mapper():
    m = mapper.Eve2CMLmapper().load()
    assert m

    d = m.as_dict()
    assert d

    with open("/dev/null", "w", encoding="utf-8") as out:
        m.dump(out)

    nd = m.node_def("iol", "iol", "i86bi_linux_l2")
    assert nd

    nd = m.node_def("qemu", "vios", "")
    assert nd

    label = m.cml_iface_label(0, "iosv", "gaga")
    assert label != "gaga"

    label = m.cml_iface_label(0, "doesntexist", "gaga")
    assert label == "gaga"


def test_custom_mapper(request, caplog):
    caplog.set_level(logging.INFO)

    caplog.clear()
    _ = mapper.Eve2CMLmapper().load("doesntexist")
    assert "mapper provided but not found" in caplog.text

    caplog.clear()
    testdata = Path(request.path).parent / "testdata" / "map1.yaml"
    _ = mapper.Eve2CMLmapper().load(str(testdata))
    assert "custom mapper loaded" in caplog.text

    caplog.clear()
    testdata = Path(request.path).parent / "testdata" / "map2.yaml"
    with pytest.raises(SystemExit) as pytest_wrapped_e:
        _ = mapper.Eve2CMLmapper().load(str(testdata))
    assert pytest_wrapped_e.type is SystemExit
    assert pytest_wrapped_e.value.code == 1
    assert "can't decode" in caplog.text

    caplog.clear()
    testdata = Path(request.path).parent / "testdata" / "hub.unl"
    with pytest.raises(SystemExit) as pytest_wrapped_e:
        _ = mapper.Eve2CMLmapper().load(str(testdata))
    assert pytest_wrapped_e.type is SystemExit
    assert pytest_wrapped_e.value.code == 1
    assert "can't use provided mapper" in caplog.text



import pytest
from eve2cml.eve.textobject import rgb_to_hex


def test_rgb_to_hex():
    assert rgb_to_hex("rgb(127, 140, 141)") == "#7F8C8DFF"
    assert rgb_to_hex("rgb(255, 0, 0)") == "#FF0000FF"
    assert rgb_to_hex("rgb(0, 255, 0)") == "#00FF00FF"
    assert rgb_to_hex("rgb(0, 0, 255)") == "#0000FFFF"
    assert rgb_to_hex("rgb(255, 255, 255)") == "#FFFFFFFF"
    assert rgb_to_hex("rgb(255, 255, 255, 0)") == "#FFFFFF00"
    assert rgb_to_hex("rgb(255, 255, 255, 128)") == "#FFFFFF80"
    assert rgb_to_hex("rgba(234, 234, 244, 0)") == "#EAEAF400"


def test_rgb_to_hex_invalid_input():
    with pytest.raises(ValueError):
        rgb_to_hex("invalid_rgb_string")
    with pytest.raises(ValueError):
        rgb_to_hex("rgb(256, 255, 255)")
    with pytest.raises(ValueError):
        rgb_to_hex("rgb(255, 255, 256)")
    with pytest.raises(ValueError):
        rgb_to_hex("rgb(-1, 0, 0)")
    with pytest.raises(ValueError):
        rgb_to_hex("rgb(0, -1, 0)")
    with pytest.raises(ValueError):
        rgb_to_hex("rgb(0, 0, -1)")



from eve2cml.eve.decode import decode_data


def test_decode_data():
    # Test when data is None
    assert decode_data(None) == ""

    # Test when data is a valid base64 encoded string
    data = "SGVsbG8gV29ybGQh"
    expected_result = "Hello World!"
    assert decode_data(data) == expected_result

    # Test when data is an invalid base64 encoded string
    invalid_data = "InvalidData"
    assert decode_data(invalid_data) == invalid_data



from typing import List
from xml.etree.ElementTree import Element


class Interface:
    def __init__(
        self,
        id: int,
        name: str,
        obj_type: str,
        network_id: int,
        labelpos="",
        curviness="",
        beziercurviness="",
        midpoint="",
        srcpos="",
        dstpos="",
        node_id=0,
        slot=0,
    ):
        self.id = id
        self.name = name
        self.obj_type = obj_type
        self.network_id = network_id

        # ignored for the moment
        self.labelpos = labelpos
        self.curviness = curviness
        self.beziercurviness = beziercurviness
        self.midpoint = midpoint
        self.srcpos = srcpos
        self.dstpos = dstpos

        # private, for linking
        self.node_id = node_id
        self.slot = slot

    def __str__(self):
        return f"ID: {self.id}, Name: {self.name}, Type: {self.obj_type}, NetID: {self.network_id}"

    def __repr__(self):
        return f"{self.__class__.__name__}(id={self.id}, slot={self.slot})"

    def as_cml_dict(self, idx, node_def, lab):
        # can likely remove the idx arg as the id should match the idx!
        assert self.id == idx
        return {
            # "id": f"i{self.id}",
            "id": f"i{idx}",
            "label": lab.mapper.cml_iface_label(self.slot, node_def, self.name),
            "slot": self.slot,
            "type": "physical",
        }

    @classmethod
    def parse(
        cls, node_id: int, obj_type: str, elem: List[Element]
    ) -> List["Interface"]:
        # special treatment for slots when type is IOL
        no_iol = obj_type != "iol"

        interfaces: List[Interface] = []
        for interface_elem in elem:
            id = int(interface_elem.attrib.get("id", "unknown"))
            interface = Interface(
                id=id,
                name=interface_elem.attrib.get("name", "unknown"),
                obj_type=interface_elem.attrib.get("type", "unknown"),
                network_id=int(interface_elem.attrib.get("network_id", 0)),
                labelpos=interface_elem.attrib.get("labelpos", ""),
                curviness=interface_elem.attrib.get("curviness", ""),
                beziercurviness=interface_elem.attrib.get("beziercurviness", ""),
                midpoint=interface_elem.attrib.get("midpoint", ""),
                srcpos=interface_elem.attrib.get("srcpos", ""),
                dstpos=interface_elem.attrib.get("dstpos", ""),
                node_id=node_id,
                slot=id if no_iol else ((id & 0xF) * 4) + (id >> 4),
            )
            interfaces.append(interface)
        return interfaces



from typing import List
from xml.etree.ElementTree import Element

from .decode import decode_data


class Config:
    def __init__(self, id: int, data: str):
        self.id = id
        self.data = data
        self._data: str

    @property
    def data(self):
        return self._data

    @data.setter
    def data(self, value: str):
        self._data = decode_data(value)

    def __str__(self):
        return f"Config ID: {self.id}, Data: {self.data}"

    @classmethod
    def parse(cls, lab: Element, path) -> List["Config"]:
        configs: List[Config] = []
        for config_elem in lab.findall(path):
            config = Config(
                id=int(config_elem.attrib.get("id", 0)), data=config_elem.text or ""
            )
            configs.append(config)
        return configs



import io
from pathlib import Path

import eve2cml.main
import pytest


@pytest.mark.parametrize(
    "case",
    [
        {
            "filename": "hub.unl",
            "nodes": 4,
            "links": 3,
            "annotations": 1,
        },
        {
            "filename": "nat.unl",
            "nodes": 4,
            "links": 2,
            "annotations": 0,
        },
        {
            "filename": "pnet.unl",
            "nodes": 4,
            "links": 3,
            "annotations": 0,
        },
        {
            "filename": "test.unl",
            "nodes": 2,
            "links": 1,
            "annotations": 4,
        },
        {
            "filename": "test.zip",
            "nodes": 2,
            "links": 1,
            "annotations": 4,
        },
    ],
)
def test_integration(request, case):
    testdata = Path(request.path).parent / "testdata" / case["filename"]
    mapper = eve2cml.main.Eve2CMLmapper().load()
    labs = eve2cml.main.convert_files(str(testdata), mapper)
    assert len(labs) == 1
    result = labs[0].as_cml_dict()
    assert len(result.keys()) == 4
    assert len(result["nodes"]) == case["nodes"]
    assert len(result["links"]) == case["links"]
    assert len(result["annotations"]) == case["annotations"]


def test_text_out(request):
    testdata = Path(request.path).parent / "testdata" / "test.unl"
    mapper = eve2cml.main.Eve2CMLmapper().load()
    labs = eve2cml.main.convert_files(str(testdata), mapper)
    assert len(labs) == 1
    text_obj = io.StringIO()
    eve2cml.main.dump_as_text(text_obj, labs[0], True)
    text_obj.seek(0)
    content = text_obj.read()
    assert len(content) > 0


def test_no_file():
    mapper = eve2cml.main.Eve2CMLmapper().load()
    with pytest.raises(SystemExit, match="1"):
        eve2cml.main.convert_files("doesntexist", mapper)



import logging

# Define color codes for different log levels
COLOR_CODES = {
    "DEBUG": "\033[36m",  # Cyan
    "INFO": "\033[32m",  # Green
    "WARNING": "\033[33m",  # Yellow
    "ERROR": "\033[31m",  # Red
    "CRITICAL": "\033[1;31m",  # Bold Red
    "RESET": "\033[0m",  # Reset
}


# Custom formatter for colorized output
class ColorFormatter(logging.Formatter):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    def format(self, record):
        log_level = record.levelname
        if log_level in COLOR_CODES:
            color_code = COLOR_CODES[log_level]
            # record.msg = f"{color_code}{record.msg}{COLOR_CODES['RESET']}"
            record.levelname = f"{color_code}{record.levelname}:{COLOR_CODES['RESET']}"

        return super().format(record)


def initialize_logging(level_str: str, nocolor: bool):
    # get the root logger
    logger = logging.getLogger()

    # Set level
    level = logging._nameToLevel.get(level_str.upper(), logging.INFO)
    logger.setLevel(level)

    # Create console handler and set level
    ch = logging.StreamHandler()
    ch.setLevel(level)

    # Create formatter and set to handler
    log_format = "%(levelname)s [%(module)s:%(lineno)d] %(message)s"
    if nocolor:
        formatter = logging.Formatter(log_format)
    else:
        formatter = ColorFormatter(log_format)
    ch.setFormatter(formatter)

    # Add handler to logger
    logger.addHandler(ch)

    # # Test logging
    # logger.debug('Debug message')
    # logger.info('Info message')
    # logger.warning('Warning message')
    # logger.error('Error message')
    # logger.critical('Critical message')



from typing import List, Optional
from xml.etree.ElementTree import Element

from .decode import decode_data


class Task:
    def __init__(self, id: int, name: str, obj_type: str, data: Optional[str] = None):
        self.id = id
        self.name = name
        self.obj_type = obj_type
        self.data = data

    def __str__(self):
        return f"Task {self.id}, Name: {self.name}, Type: {self.obj_type}, Data: {decode_data(self.data)}"
        # return f"Task {self.id}, Name: {self.name}, Type: {self.obj_type}"

    @classmethod
    def parse(cls, lab: Element, path: str) -> List["Task"]:
        tasks: List[Task] = []
        for task_elem in lab.findall(path):
            task = Task(
                id=int(task_elem.attrib.get("id", 0)),
                name=task_elem.attrib.get("name", "unknown"),
                obj_type=task_elem.attrib.get("type", "unknown"),
            )
            data = task_elem.find("data")
            if data is not None and data.text:
                task.data = data.text
            tasks.append(task)
        return tasks



import base64
import logging
from typing import Optional

_LOGGER = logging.getLogger(__name__)


def decode_data(data: Optional[str]) -> str:
    if data is None:
        return ""
    try:
        decoded = base64.b64decode(data).decode("utf-8")
    except Exception as exc:
        _LOGGER.exception("b64 decode %s", exc)
        decoded = data
    return decoded



import logging
from typing import Any, Dict, List
from xml.etree.ElementTree import Element

from .config import Config
from .configset import ConfigSet
from .task import Task
from .textobject import TextObject

_LOGGER = logging.getLogger(__name__)


class Objects:
    def __init__(
        self,
        tasks: List[Task],
        configs: List[Config],
        configsets: List[ConfigSet],
        textobjects: List[TextObject],
    ):
        self.tasks = tasks
        self.configs = configs
        self.configsets = configsets
        self.textobjects = textobjects

    def __str__(self):
        return f"Tasks: {self.tasks}, Configs: {self.configs}, Config Sets: {self.configsets}, Text Objects: {self.textobjects}"

    def get_config(self, cfg_set: int, id: int):
        if cfg_set == 1:
            for config in self.configs:
                if config.id == id:
                    return config.data
            return ""
        for config_set in self.configsets:
            for config in config_set.configs:
                if config.id == id:
                    return config.data
        return ""

    @classmethod
    def parse(cls, lab: Element, path: str, filename: str) -> "Objects":
        objects = lab.findall(path)
        if len(objects) == 0:
            return Objects(tasks=[], configs=[], configsets=[], textobjects=[])
        if len(objects) > 1:
            _LOGGER.info(
                "more than one object in tree (%d) for %s", len(objects), filename
            )
        this = objects[0]
        return Objects(
            Task.parse(this, "tasks/task"),
            Config.parse(this, "configs/config"),
            ConfigSet.parse(this, "configsets/configset"),
            TextObject.parse(this, "textobjects/textobject"),
        )

    def cml_annotations(self) -> List[Dict[str, Any]]:
        annotations: List[Dict[str, Any]] = []
        for object in self.textobjects:
            annotation_list = object.as_cml_annotations()
            for annotation in annotation_list:
                annotations.append(annotation)
        return annotations



import logging

import pytest
from eve2cml.log import initialize_logging


@pytest.mark.parametrize(
    "nocolor",
    [True, False],
)
def test_initialize_logging(caplog, nocolor):
    initialize_logging("DEBUG", nocolor=nocolor)

    # Verify log messages are colorized
    logger = logging.getLogger()
    logger.debug("Debug message")
    logger.info("Info message")
    logger.warning("Warning message")
    logger.error("Error message")
    logger.critical("Critical message")

    # Retrieve captured log records
    log_records = caplog.records

    # Verify colored log levels
    if not nocolor:
        for record in log_records:
            assert record.levelname.startswith("\033[")
            assert record.levelname.endswith("\033[0m")

    # Verify log messages content
    assert log_records[0].message == "Debug message"
    assert log_records[1].message == "Info message"
    assert log_records[2].message == "Warning message"
    assert log_records[3].message == "Error message"
    assert log_records[4].message == "Critical message"



import io
import logging
import sys
from importlib.resources import files
from pathlib import Path
from typing import Dict, List, Optional

import yaml
from yaml.parser import ParserError

_LOGGER = logging.getLogger(__name__)


class CMLdef:
    def __init__(
        self, node_def: str, image_def: Optional[str] = None, override: bool = False
    ):
        self.node_def = node_def
        self.image_def = image_def
        self.override = override

    def __repr__(self):
        return f"{self.__class__.__name__}(node_def={self.node_def}, image_def={self.image_def}, override={self.override})"

    def as_dict(self):
        return {
            "node_def": self.node_def,
            "image_def": self.image_def,
            "override": self.override,
        }


class Eve2CMLmapper:
    def __init__(self):
        self.map: Dict[str, CMLdef] = {}
        self.unknown_type: str = ""
        self.interface_lists: Dict[str, List[str]] = {}

    def as_dict(self):
        return {
            "map": {k: v.as_dict() for k, v in self.map.items()},
            "unknown_type": self.unknown_type,
            "interface_lists": self.interface_lists,
        }

    def dump(self, out: io.TextIOWrapper):
        return yaml.safe_dump(self.as_dict(), out, indent=2)

    @classmethod
    def load(cls, filename="") -> "Eve2CMLmapper":
        map_data = yaml.safe_load(
            files("eve2cml.map_data").joinpath("default.yaml").read_text()
        )

        if filename:
            map_file = Path(filename)
            if map_file.is_file():
                with open(map_file) as fh:
                    try:
                        map_data = yaml.safe_load(fh)
                    except ParserError as exc:
                        _LOGGER.critical("can't decode %s: %s", filename, exc)
                        sys.exit(1)
                if not isinstance(map_data, dict):
                    _LOGGER.critical("can't use provided mapper file")
                    sys.exit(1)
                _LOGGER.warning("custom mapper loaded: %s", filename)
            else:
                _LOGGER.error("mapper provided but not found. Using built-in mapper!")

        mapper = Eve2CMLmapper()
        mapper.unknown_type = map_data.get("unknown_type", "")
        mapper.interface_lists = map_data.get("interface_lists", {})
        for key, value in map_data.get("map", {}).items():
            mapper.map[key] = CMLdef(**value)

        return mapper

    def node_def(self, obj_type: str, template: str, image: str) -> CMLdef:
        lookup = f"{obj_type}:{template}"
        if len(image) > 0:
            lookup += f"{lookup}:{image}"
        found = self.map.get(lookup)
        if not found:
            # special case for non-template images like IOL or Docker
            longest_prefix = ""
            longest_cmldef = None
            for key, cmldef in self.map.items():
                if lookup.startswith(key) and len(key) > len(longest_prefix):
                    longest_prefix = key
                    longest_cmldef = cmldef
            if longest_cmldef:
                _LOGGER.info("mapped node type %s", longest_cmldef)
                return longest_cmldef
            _LOGGER.warning("Unmapped node type %s %s %s", obj_type, template, image)
            return CMLdef(self.unknown_type, None, True)
        return found

    def cml_iface_label(self, slot: int, node_def: str, label: str) -> str:
        interfaces = self.interface_lists.get(node_def)
        if interfaces is None:
            _LOGGER.warning("No mapping: %s %d", node_def, slot)
            return label
        return interfaces[slot]



import logging
import re
from functools import cached_property
from typing import Any, Dict, List
from xml.etree.ElementTree import Element

from bs4 import BeautifulSoup

from .decode import decode_data

_LOGGER = logging.getLogger(__name__)

# just a default color as a fallback
GRAY = "#808080FF"


def parse_style(style_string):
    style_dict = {}
    style_pairs = style_string.split(";")
    for pair in style_pairs:
        if pair.strip():  # Skip empty strings
            key, value = pair.split(":")
            style_dict[key.strip()] = value.strip()
    # _LOGGER.info("style %s %s", style_string, style_dict)
    return style_dict


def rgb_to_hex(rgb_string):
    # Split the RGB components and convert them to integers
    exp = r"rgba?\((\d{1,3}),\s*(\d{1,3}),\s*(\d{1,3})(?:,\s*(\d{1,3}))?\)"
    match = re.match(exp, rgb_string)
    if match:
        values = [255 if v is None else int(v) for v in match.groups()]
        for value in values:
            if value < 0 or value > 255:
                raise ValueError
        return f'#{"".join([f"{int(v):02x}" for v in values])}'.upper()
    raise ValueError


def color_convert(color: str) -> str:
    if color.startswith("rgb"):
        color = rgb_to_hex(color)
    if color.startswith("#") and len(color) == 7:
        return f"{color}FF"
    return color


def parse_rotate(rotate: str) -> int:
    exp = r"rotate\((\d{1,3})deg\)"
    match = re.match(exp, rotate)
    if match:
        return int(match.group(1))
    return 0


class TextObject:
    def __init__(self, id: int, name: str, obj_type: str, data=""):
        self.id = id
        self.name = name
        self.obj_type = obj_type
        self._data = None
        self._div = None
        self._div_style = None
        if data:
            self.data = data

    @property
    def data(self):
        return self._data

    @data.setter
    def data(self, value: str):
        self._data = BeautifulSoup(decode_data(value), "html.parser")
        self._div = self._data.find_all("div", class_="customShape")
        if len(self._div) > 0:
            self._div_style = parse_style(self._div[0]["style"])

    @cached_property
    def style_summary(self):
        def has_style(tag):
            has = tag.has_attr("style")
            return has

        if self._data is None or self._data.div is None:
            return {}

        style = self._data.div.find_all(has_style)
        styles: Dict[str, str] = {}
        for el in style:
            el_style = parse_style(el["style"])
            styles = {**styles, **el_style}

        # Check out potential font tags for color information
        # This is quite a hack as there could be different colors in a text
        # object, also font sizes and what else...  Pretty much the Wild West.
        font_tags = self._data.div.find_all("font")
        if len(font_tags) > 0:
            # First color in list? Or the last... Guessing
            color = font_tags[0].attrs.get("color")
            if color is not None:
                styles["color"] = color
        return styles

    def prettify(self) -> str:
        if self.data is not None:
            return self.data.prettify()
        return ""

    @property
    def strings(self) -> str:
        if self.data is not None:
            return "\n".join(self.data.stripped_strings)
        return ""

    @property
    def left(self) -> int:
        if self._div_style is None:
            return 0
        return int(float(self._div_style["left"].strip("px")))

    @property
    def top(self) -> int:
        if self._div_style is None:
            return 0
        return int(float(self._div_style["top"].strip("px")))

    @property
    def width(self) -> int:
        if self._div_style is None:
            return 0
        value = self._div_style.get("width", "")
        # best guesses
        if value == "" or value == "auto":
            line_length = max([len(line) for line in self.strings.split("\n")])
            return int(
                line_length
                * 0.67
                * float(self.style_summary.get("font-size", "14").strip("px"))
            )
        return int(float(value.strip("px")))

    @property
    def height(self) -> int:
        if self._div_style is None:
            return 0
        value = self._div_style.get("height", "")
        # best guesses
        if value == "" or value == "auto":
            lines = self.strings.rstrip().count("\n") + 1
            return int(
                lines
                * 1.5
                * float(self.style_summary.get("font-size", "14").strip("px"))
            )
        return int(float(value.strip("px")))

    @property
    def z_index(self) -> int:
        if self._div_style is None:
            return 0
        return int(self._div_style.get("z-index", 0))

    @property
    def rotation(self) -> int:
        if self._div_style is None:
            return 0
        transform = self._div_style.get("transform")
        if transform is None:
            return 0
        return int(parse_rotate(transform))

    def __str__(self) -> str:
        return f"Text ID: {self.id}, Name: {self.name}, Type: {self.obj_type}, Strings: {self.strings}, Data: {self.prettify()}, Pos: {self.left}/{self.top}/{self.z_index}"

    @classmethod
    def parse(cls, lab: Element, path: str) -> List["TextObject"]:
        text_objects: List[TextObject] = []
        for text_elem in lab.findall(path):
            text_object = TextObject(
                id=int(text_elem.attrib.get("id", 0)),
                name=text_elem.attrib.get("name") or "",
                obj_type=text_elem.attrib.get("type") or "",
            )
            data = text_elem.find("data")
            if data is not None and data.text:
                text_object.data = data.text
            text_objects.append(text_object)
        return text_objects

    def as_cml_annotations(self) -> List[Dict[str, Any]]:
        if self.obj_type == "text":
            if not len(self.strings) > 0:
                _LOGGER.info("Not a real text object, ignoring")
                return []

            _LOGGER.info("Procssing ID %d, TEXT", self.id)
            color = color_convert(self.style_summary.get("color", GRAY))
            text_size = self.style_summary.get("font-size", "16px")
            background_color = self.style_summary.get("background-color", "")
            if background_color:
                background_color = color_convert(background_color)
            font_family = self.style_summary.get("font-family", "serif")
            ret_val = [
                {
                    "border_color": "#FFFFFF00",
                    "border_style": "",
                    "color": color,
                    "rotation": self.rotation,
                    "text_bold": False,
                    "text_content": self.strings,
                    "text_font": font_family,
                    "text_italic": False,
                    "text_size": max(int(float(text_size.strip("px"))), 8),
                    "text_unit": "pt",
                    "thickness": 1,
                    "type": "text",
                    "x1": self.left,
                    "y1": self.top,
                    "z_index": self.z_index,
                }
            ]
            if background_color != "":
                ret_val.append(
                    {
                        "border_color": "#FFFFFF00",
                        "border_radius": 0,
                        "border_style": "",
                        "color": background_color,
                        # can't rotate a rectangle!!!
                        # "rotation": self.rotation,
                        "thickness": 1,
                        "type": "rectangle",
                        "x1": self.left,
                        "y1": self.top,
                        "x2": self.width,
                        "y2": self.height,
                        "z_index": self.z_index - 1,
                    }
                )
            return ret_val

        elif self.obj_type == "square":
            _LOGGER.info("Procssing ID %d, SQUARE", self.id)
            summary = {**self.style_summary, **self._data.div.svg.rect.attrs}  # type: ignore
            return [
                {
                    "border_color": color_convert(summary.get("stroke", GRAY)),
                    "border_radius": int(float(summary.get("rx", "0"))),
                    "border_style": summary.get("bla", ""),
                    "color": color_convert(summary.get("fill", GRAY)),
                    "thickness": max(int(float(summary.get("stroke-width", "1"))), 1),
                    "type": "rectangle",
                    "x1": self.left,
                    "y1": self.top,
                    "x2": float(summary.get("width", "80")),
                    "y2": float(summary.get("height", "80")),
                    "z_index": self.z_index,
                }
            ]

        elif self.obj_type == "circle":
            _LOGGER.info("Procssing ID %d, CIRCLE", self.id)
            summary = {**self.style_summary, **self._data.div.svg.ellipse.attrs}  # type: ignore
            rx = float(summary.get("rx", "80"))
            ry = float(summary.get("ry", "80"))
            return [
                {
                    "border_color": color_convert(summary.get("stroke", GRAY)),
                    "border_style": summary.get("bla", ""),
                    "color": color_convert(summary.get("stroke", GRAY)),
                    "thickness": max(int(float(summary.get("stroke-width", "1"))), 1),
                    "type": "ellipse",
                    "x1": rx + self.left,
                    "y1": ry + self.top,
                    "x2": rx,
                    "y2": ry,
                    "z_index": self.z_index,
                }
            ]
        else:
            _LOGGER.warn("Object type %s", self.obj_type)
        return []



import logging
from typing import TYPE_CHECKING, List
from xml.etree.ElementTree import Element

from . import Interface

if TYPE_CHECKING:
    from .lab import Lab

_LOGGER = logging.getLogger(__name__)


class Node:
    def __init__(
        self,
        id: int,
        name: str,
        interfaces: List[Interface],
        obj_type="",
        template="",
        image="",
        console="",
        cpu=0,
        cpulimit=0,
        ram=0,
        ethernet=0,
        uuid="",
        firstmac="",
        qemu_options="",
        qemu_version="",
        qemu_arch="",
        delay=0,
        sat=0,
        icon="",
        config=0,
        left=0,
        top=0,
        e0dhcp="",
    ):
        self.id = id
        self.name = name
        self.interfaces = interfaces
        self.obj_type = obj_type
        self.template = template
        self.image = image
        self.console = console
        self.cpu = cpu
        self.cpulimit = cpulimit
        self.ram = ram
        self.ethernet = ethernet
        self.uuid = uuid
        self.firstmac = firstmac
        self.qemu_options = qemu_options
        self.qemu_version = qemu_version
        self.qemu_arch = qemu_arch
        self.delay = delay
        self.sat = sat
        self.icon = icon
        self.config = config
        self.left = left
        self.top = top
        self.e0dhcp = e0dhcp

        self.cml_hide_links = False
        self.cml_config = ""

    def __str__(self):
        return f"ID: {self.id}, Name: {self.name}, Type: {self.obj_type}, X: {self.left}, Y: {self.top}, Template: {self.template}, Image: {self.image}, Ethernet: {self.ethernet}"

    def as_cml_dict(self, node_id: int, lab: "Lab"):
        nd_map = lab.mapper.node_def(self.obj_type, self.template, self.image)

        temp_list: List[Interface] = []
        prev_idx = 0
        prev_slot = 0
        _LOGGER.debug(self.interfaces)
        for idx, iface in enumerate(self.interfaces):
            delta = iface.slot - prev_slot
            _LOGGER.debug(
                "idx, slot, prev, delta %d/%d/%d/%d", idx, iface.slot, prev_slot, delta
            )
            for idx2 in range(delta):
                _LOGGER.debug(
                    "prepending filler interface id %d/%d", prev_idx, prev_slot + idx2
                )
                temp_list.append(
                    Interface(
                        id=prev_idx,
                        obj_type=self.obj_type,
                        network_id=999999,
                        name="filler",
                        slot=prev_slot + idx2,
                    )
                )
                prev_idx += 1
            iface.id = prev_idx
            temp_list.append(iface)
            prev_slot = iface.slot + 1
            prev_idx += 1

        _LOGGER.debug("list %s", temp_list)

        iface_count = len(temp_list)
        iface_diff = int(self.ethernet) - iface_count
        if iface_diff > 0:
            _LOGGER.info(
                "Filler interfaces needed for Node %d/%s, add %d",
                self.id,
                self.name,
                iface_diff,
            )
            for iface_idx in range(iface_diff):
                id = iface_count + iface_idx
                temp_list.append(
                    Interface(
                        id=id,
                        obj_type=self.obj_type,
                        network_id=999999,
                        slot=id,
                        name="doesntmatterhere",
                    )
                )

        # If there's no config then use the cml config (which also might be the
        # empty string)
        config = lab.objects.get_config(self.config, self.id) or self.cml_config

        _LOGGER.info("Serializing %s %s", self.name, self.id)
        return {
            "id": f"n{node_id}",
            "boot_disk_size": None,
            "configuration": config,
            "cpu_limit": 100 - int(self.cpulimit) if self.cpulimit else None,
            "cpus": int(self.cpu) if (self.cpu and not nd_map.override) else None,
            "data_volume": None,
            "hide_links": self.cml_hide_links,
            "label": self.name,
            "node_definition": nd_map.node_def,
            "image_definition": nd_map.image_def,
            "ram": int(self.ram) if (self.ram and not nd_map.override) else None,
            "tags": [],
            "x": int(self.left),
            "y": int(self.top),
            "interfaces": [
                iface.as_cml_dict(idx, nd_map.node_def, lab)
                for idx, iface in enumerate(temp_list)
            ],
        }

    @classmethod
    def parse(cls, lab: Element) -> List["Node"]:
        nodes: List[Node] = []
        for node_elem in lab.findall(".//node"):
            id = int(node_elem.attrib.get("id", 0))
            obj_type = node_elem.attrib.get("type", "unknown")
            node = Node(
                id=id,
                name=node_elem.attrib.get("name", "unknown"),
                interfaces=Interface.parse(
                    id, obj_type, node_elem.findall("interface")
                ),
                obj_type=obj_type,
                template=node_elem.attrib.get("template", "unknown"),
                image=node_elem.attrib.get("image", "unknown"),
                console=node_elem.attrib.get("console", "unknown"),
                cpu=int(node_elem.attrib.get("cpu", 0)),
                cpulimit=int(node_elem.attrib.get("cpulimit", 0)),
                ram=int(node_elem.attrib.get("ram", 0)),
                ethernet=int(node_elem.attrib.get("ethernet", 0)),
                uuid=node_elem.attrib.get("uuid", ""),
                firstmac=node_elem.attrib.get("firstmac", ""),
                qemu_options=node_elem.attrib.get("qemu_options", ""),
                qemu_version=node_elem.attrib.get("qemu_version", ""),
                qemu_arch=node_elem.attrib.get("qemu_arch", ""),
                delay=int(node_elem.attrib.get("delay", 0)),
                sat=int(node_elem.attrib.get("sat", 0)),
                icon=node_elem.attrib.get("icon", ""),
                config=int(node_elem.attrib.get("config", 0)),
                left=int(node_elem.attrib.get("left", 0)),
                top=int(node_elem.attrib.get("top", 0)),
                e0dhcp=node_elem.attrib.get("e0dhcp", ""),
            )
            nodes.append(node)
        return nodes
